C
C     This program reads an input file of flashcard specifications
C     and generates an output file that can be processed like any
C     document containing Hebrew <transliterations>.
C
C     usage:
C     cat in.flash | ${HOME}/bin/exe/flashcards ["title"] > out.heb
C
C     unit  purpose
C     ----  -------
C       0   prompts and error messages
C       1   ${HOME}/Utility/hebrew.hsh containing the dictionary
C       5   flashcard specifications
C       6   output .heb file
C
C     exit  meaning
C     ----  -------
C       0   all went well
C       1   an input line was not properly formatted
C       2   a transliteration was not found in the dictionary
C       3   more than one command-line parameter was supplied
C
C     variable  meaning
C     --------  -------
C     BNK18     a string of 16 blanks
C     BNK80     a string of 80 blanks
C     CBNK80    BNK80 as separate characters
C     DRAW      routine generates LaTeX commands to print flashcards
C     ENGLS     English translations found in the dictionary
C     ENGS      English for 3 cards each having up to MXON words
C     EXIT      system routine stops the program with a return code
C     GETENG    routine tables English translations of Hebrew words
C     GETHSH    routine tables lenfths and hashcodes of Hebrew words
C     I         index on the 3 flashcards of the current page
C     IN        T => copying a transliteration
C     J         index on the NW(I) words of card I
C     K         index on the characters of a string
C     L         length of an input line
C     LENGTH    function returns index of last nonblank in a string
C     LETRS     letter hash codes of Hebrew words in dictionary
C     LINE      input line with all the transliterations for a card
C     LNUM      number of input line just read
C     LTRS      letter hash codes for 3 cards each up to MXON words
C     M         dictionary index of a transliteration
C     MXCH      maximum characters allowed in a Hebrew word
C     MXON      maximum words on a flashcard
C     MXWD      maximum dictionary words allowed
C     NC        character counts for 3 cards each up to MXON words
C     NCS       letter counts of Hebrew words in the dictionary
C     NW(I)     number of words on card I
C     NWD       number of words found in the dictionary
C     RC        return code from SHFTKN; 0 => all went well
C     SHFTKN    routine shifts string left by given number of tokens
C     SPACEL    routine removes leading and extra blanks
C     TITLE     routine writes LaTeX for a title page
C     TRANS     Hebrew transliterations found in the dictionary
C     TRN       a single transliteration as one long word
C     TRNC      TRN as separate characters
C     TRNIDX    function gives index of <transliteration> in a list
C     TRNS      transliterations for 3 cards each up to MXON words
C     VEES      vowel hash codes for 3 cards each up to MXON words
C     VS        vowel hash codes of Hebrew dictionary words
C
C     prepare to store the dictionary
      PARAMETER(MXWD=6000)
      CHARACTER*80 ENGLS(MXWD)
      CHARACTER*18 TRANS(MXWD)
      PARAMETER(MXCH=12)
      INTEGER*1 NCS(MXWD),LETRS(MXCH,MXWD),VS(MXCH,MXWD)
C
C     prepare to read input lines
      CHARACTER*1 LINE(80)
C
C     prepare to extract a single transliteration
      LOGICAL*4 IN
      CHARACTER*18 TRN
      CHARACTER*1 TRNC(18)
      EQUIVALENCE(TRNC,TRN)
C
C     prepare to store the transliterations for three cards
      PARAMETER(MXON=7)
      CHARACTER*18 TRNS(3,MXON)
      CHARACTER*80 ENGS(3,MXON)
      INTEGER*1 NC(3,MXON),LTRS(MXCH,3,MXON),VEES(MXCH,3,MXON)
      INTEGER*4 NW(3)
C
C     other local variables
      INTEGER*4 TRNIDX,RC
      CHARACTER*18 BNK18/'                  '/
      CHARACTER*80 BNK80
      CHARACTER*1 CBNK80(80)/80*' '/
      EQUIVALENCE(BNK80,CBNK80)
C
C ------------------------------------------------------------------
C
C     table Hebrew transliterations and English translations
      CALL GETENG(     MXWD, NWD,    TRANS,   ENGLS)
      CALL GETHSH(MXCH,MXWD, NWD,NCS,TRANS,VS,LETRS)
C
C     write the preamble in the output file
      WRITE(6,901)
  901 FORMAT('\documentclass[12pt]{article}'/
     ;       '\input{${HOME}/bin/hebrew}'/
     ;       '\usepackage{fullpage}'/
     ;       '\pagestyle{empty}'/
     ;       '\usepackage{float}'/
     ;       '\begin{document}'/
     ;       '\setivrit{17}')
C
C     if a title is specified, write a title page
      CALL TITLE
C
C     prepare to count input lines
      LNUM=0
C
C     read the transliterations for the next page of 3 flashcards
   14 I=0
C
C     initialize for this flashcard
   13 I=I+1
      DO 1 J=1,MXON
           TRNS(I,J)=BNK18
           ENGS(I,J)=BNK80
           NC(I,J)=0
           DO 2 K=1,MXCH
                LTRS(K,I,J)=0
                VEES(K,I,J)=0
    2      CONTINUE
    1 CONTINUE
C
C     read the transliterations for this flashcard
    8 READ(5,902,END=3) LINE
  902 FORMAT(80A1)
      LNUM=LNUM+1
      CALL SPACEL(LINE,80)
      L=LENGTH(LINE,80)
      NW(I)=0
C     a blank line denotes a blank flashcard
      IF(L.EQ.0) GO TO 4
C     skip comment lines
      IF(LINE(1).EQ.'%') GO TO 5
C     ignore anything following the last > on the line
      DO 6 K=L,1,-1
           IF(LINE(K).EQ.'>') GO TO 7
           LINE(K)=' '
    6 CONTINUE
    5 WRITE(0,991) LNUM
  991 FORMAT('skipping nonblank line ',I3)
      GO TO 8
C
C     extract the next transliteration from the input line
    7 L=LENGTH(LINE,L)
   12 NW(I)=NW(I)+1
      TRN=BNK18
      K=0
      IN=.FALSE.
    9 K=K+1
      IF(.NOT.IN) THEN
C        out of transliteration
         IF(LINE(K).EQ.' ') GO TO 9
         IF(LINE(K).EQ.'<') THEN
            IN=.TRUE.
            GO TO 9
         ELSE
            WRITE(0,905) LINE(K),LNUM,K,NW(I)-1
  905       FORMAT('text ',A1,' outside transliteration on line',
     ;             I3,' at column ',I2,' after token ',I1)
            CALL EXIT(1)
         ENDIF
      ELSE
C        in transliteration
         IF(LINE(K).EQ.'>') THEN
            IN=.FALSE.
            GO TO 10
         ELSE
            TRNC(K)=LINE(K)
            GO TO 9
         ENDIF
      ENDIF
C
C     print a blank line for a blank transliteration
   10 IF(TRN.EQ.BNK18) THEN
         TRNS(I,NW(I))=BNK18
         ENGS(I,NW(I))=BNK80
         GO TO 21
      ENDIF
C
C     find the transliteration in the table
      CALL SPACEL(TRNC,18)
      M=TRNIDX(TRN,TRANS,NWD)
      IF(M.EQ.0) THEN
C        report that the transliteration was not found
         WRITE(0,903) NW(I),LNUM,TRN
  903    FORMAT('transliteration ',I1,' on line ',I3,
     ;          ' not found: ',A18)
         CALL EXIT(2)
      ELSE
C        remember it and the corresponding translation
         TRNS(I,NW(I))=TRN
         ENGS(I,NW(I))=ENGLS(M)
         DO 11 K=1,NCS(M)
               LTRS(K,I,NW(I))=LETRS(K,M)
               VEES(K,I,NW(I))=VS(K,M)
   11    CONTINUE
         NC(I,NW(I))=NCS(M)
      ENDIF
C
C     discard this transliteration and go for the next
   21 CALL SHFTKN(LINE,L,1,RC)
      L=LENGTH(LINE,L)
      IF(L.GT.0) THEN
         IF(NW(I).LT.MXON) GO TO 12
         GO TO 4
      ENDIF
    4 IF(I.LT.3) GO TO 13
C
C     generate picture commands for the three cards on this page
      CALL DRAW(MXON,NW,TRNS,ENGS,NC,MXCH,LTRS,VEES)
C
C     advance to the next page
      GO TO 14
C
C     end of input file; complete the last page of output
    3 I=I-1
      IF(I.LT.3) NW(3)=0
      IF(I.LT.2) NW(2)=0
      IF(I.LT.1) NW(1)=0
      IF(NW(1).GT.0) CALL DRAW(MXON,NW,TRNS,ENGS,NC,MXCH,LTRS,VEES)
C
C     finish the output file
      WRITE(6,904)
  904 FORMAT('\end{document}')
      CALL EXIT(0)
      END
C
      SUBROUTINE DRAW(MXON,NW,TRNS,ENGS,NC,MXCH,LTRS,VEES)
C     This routine draws the fronts and backs of the three cards.
C
C     variable  meaning
C     --------  -------
C     BTD       routine converts an INTEGER*4 to numerals
C     CTR       height on the page of the center of a flashcard
C     E         width of an English translation
C     ENG       a single translation as one long word
C     ENGS      English for 3 cards each having up to MXON words
C     FILENO    the filing order of a flashcard
C     FLOAT     Fortran function returns REAL*4 for INTEGER*4
C     H         width of a Hebrew string
C     HEBWID    function returns printed width of a Hebrew string
C     I         index on the cards of this page
C     IFIX      Fortran function returns INTEGER*4 for REAL*4
C     INTINS    routine inserts an integer in a character string
C     J         index on the words of card I
C     K         index on the letters of a word
C     L         length of a transliteration or translation
C     LE        index of last nonblank in ENG
C     LENGTH    function returns index of last nonblank in a string
C     LHT       height of Hebrew+interline space
C     LSPC      vertical space between Hebrew words and translations
C     LT        index of last nonblank in TRN
C     LTR       a letter hash code as INTEGER*4
C     LTRS      letter hash codes of Hebrew words on this page
C     MXCH      maximum letters in a Hebrew word
C     MXON      maximum words on a flashcard
C     NC        numbers of characters in Hebrew words
C     NW(I)     number of words on card I
C     PACKL     routine removes leading and embedded blanks
C     PTR       pointer to the next hash code in FILENO
C     PTS       points of size that makes English 361 points wide
C     PUT       a \put command
C     PUTE      \put(           ,           ){\large ...
C     PUTN      \put(           ,           ){\normalsize ...
C     PUTF      \put(           ,           ){\footnotesize ...
C     PUTI      \put(           ,           ){\tiny\tt{...80*' '...}
C     PUTT      \put(           ,           ){<...18*' '...>}
C     RC        return code from BTD; 0 => all went well
C     SHIFTL    routine removes leading blanks from a string
C     SHT       height of the stack of words
C     SPACEL    routine removes leading and extra blanks
C     STRBNK    routine blanks out a character string
C     STRINS    routine inserts one character string into another
C     STRWID    function returns width of string when printed
C     TRN       a single transliteration as one long word
C     TRNS      transliterations 3 cards each with up to MXON words
C     VEES      vowel and trope hash codes
C     W         width of a Hebrew word
C     WPP       points of English width per point of size
C     X         horizontal position of a put
C     Y         vertical position of a put
C     YC        vertical position of a framebox
C
C     formal parameters
      INTEGER*4 NW(3)
      CHARACTER*18 TRNS(3,MXON)
      CHARACTER*80 ENGS(3,MXON)
      INTEGER*1 NC(3,MXON),LTRS(MXCH,3,MXON),VEES(MXCH,3,MXON)
C
C     formats for \put commands
      CHARACTER*1 PUT(125)
      CHARACTER*1 PUTE(125)/'\','p','u','t','(',' ',' ',' ',' ',' ',
     ;                      ' ',' ',' ',' ',' ',' ',',',' ',' ',' ',
     ;                      ' ',' ',' ',' ',' ',' ',' ',' ',')','{',
     ;                      '\','l','a','r','g','e',' ',80*' ','}',
     ;                      7*' '/
      CHARACTER*1 PUTN(125)/'\','p','u','t','(',' ',' ',' ',' ',' ',
     ;                      ' ',' ',' ',' ',' ',' ',',',' ',' ',' ',
     ;                      ' ',' ',' ',' ',' ',' ',' ',' ',')','{',
     ;                      '\','n','o','r','m','a','l','s','i','z',
     ;                      'e',' ',80*' ','}',2*' '/
      CHARACTER*1 PUTF(125)/'\','p','u','t','(',' ',' ',' ',' ',' ',
     ;                      ' ',' ',' ',' ',' ',' ',',',' ',' ',' ',
     ;                      ' ',' ',' ',' ',' ',' ',' ',' ',')','{',
     ;                      '\','f','o','o','t','n','o','t','e','s',
     ;                      'i','z','e',' ',80*' ','}'/
      CHARACTER*1 PUTI(125)/'\','p','u','t','(',' ',' ',' ',' ',' ',
     ;                      ' ',' ',' ',' ',' ',' ',',',' ',' ',' ',
     ;                      ' ',' ',' ',' ',' ',' ',' ',' ',')','{',
     ;                      '\','t','i','n','y','\','t','t','{',
     ;                      80*' ','}','}',4*' '/
      CHARACTER*1 PUTT(49)/'\','p','u','t','(',' ',' ',' ',' ',' ',
     ;                     ' ',' ',' ',' ',' ',' ',',',' ',' ',' ',
     ;                     ' ',' ',' ',' ',' ',' ',' ',' ',')','{',
     ;                     '<',' ',' ',' ',' ',' ',' ',' ',' ',' ',
     ;                     ' ',' ',' ',' ',' ',' ',' ','>','}'/
C
C     other local variables
      CHARACTER*1 FILENO(80)
      CHARACTER*80 ENG
      CHARACTER*18 TRN
      INTEGER*4 W,X,Y,YC,PTR,RC
      REAL*4 LHT,LSPC/6.83/
C
C ------------------------------------------------------------------
C
C     draw the fronts of the three cards on this page
      LHT=17.17+LSPC
      WRITE(6,901)
  901 FORMAT('\begin{figure}[H]'/
     ;       '\centering'/
     ;       '\begin{picture}(361,651)')
      DO 1 I=1,3
           YC=434-217*(I-1)
           WRITE(6,902) YC
  902      FORMAT('\put(0,',I3.3,'){\framebox(361,216){}}%')
C
C          find the height of the center of this card
           CTR=217.0/2.0+217.0*FLOAT(3-I)
C
C          find hight of text at 17.17pt each word + LSPC pt between
           SHT=17.17*FLOAT(NW(I))+LSPC*FLOAT((NW(I)-1))
C
C          find the height of the top of the stack of words
           Y=IFIX(CTR+SHT/2.+LHT/2.)
C          plus half the height of a line for subtraction first
C
           CALL STRBNK(FILENO,80)
           PTR=1
           DO 2 J=1,NW(I)
C               the next transliteration is LHT points lower
                Y=Y-LHT
C
C               extract next transliteration from list for this card
                TRN=TRNS(I,J)
                LT=LENGTH(TRN,18)
                IF(LT.EQ.0) GO TO 2
C
C               figure out where to put the Hebrew on the page
                H=HEBWID(LTRS(1,I,J),VEES(1,I,J),NC(I,J),17.0)
                W=IFIX(0.5+H)
                X=(361-W)/2
C
C               put coordinates and <word> in the output line
                CALL STRBNK(PUT,125)
                CALL INTINS(PUTT,49,X,6,0, PUT,L)
                CALL INTINS(PUT,125,Y,18,0, PUT,L)
                CALL STRINS(PUT,125,TRN,LT,32,0, PUT,L)
C
C               remove extra blanks and write the output line
                CALL PACKL(PUT,125)
                L=LENGTH(PUT,125)
                WRITE(6,903) (PUT(K),K=1,L)
  903           FORMAT(125A1)
C
C               accumulate this card's file number
                DO 3 K=1,NC(I,J)
                     LTR=LTRS(K,I,J)
                     CALL BTD(LTR,FILENO(PTR),2,RC)
                     IF(FILENO(PTR).EQ.' ') FILENO(PTR)='0'
                     PTR=PTR+2
                     IF(PTR.GE.79) GO TO 2
    3           CONTINUE
    2      CONTINUE
C
C          write this card's file number
           CALL STRBNK(PUT,125)
           CALL INTINS(PUTI,125,2,6,0, PUT,L)
           CALL INTINS(PUT,125,YC+209,18,0, PUT,L)
           CALL STRINS(PUT,125,FILENO,PTR-1,40,0, PUT,L)
           CALL PACKL(PUT,125)
           L=LENGTH(PUT,125)
           WRITE(6,903) (PUT(K),K=1,L)
    1 CONTINUE
C
C     end the front page
      WRITE(6,904)
  904 FORMAT('\end{picture}'/
     ;       '\vspace{-100pt}'/
     ;       '\end{figure}'
     ;       '\newpage')
C
C     draw the backs of the three cards on this page
      WRITE(6,901)
      DO 4 I=1,3
C          find the height of the center of this card
           CTR=217.0/2.0+217.0*FLOAT(3-I)
C
C          find the hight of the text to align with the Hebrew
           SHT=17.17*FLOAT(NW(I))+LSPC*FLOAT((NW(I)-1))
C
C          find the height of the top of the stack of words
           Y=IFIX(CTR+SHT/2.+LHT/2.)
C          plus half the height of a line for subtraction first
C
           DO 9 J=1,NW(I)
C               the next translation goes LHT points down
                Y=Y-LHT
C
C               extract next transliteration from list for this card
                ENG=ENGS(I,J)
                LE=LENGTH(ENG,80)
                IF(LE.EQ.0) GO TO 9
C
C               width of English text at 12 point \large
                E=STRWID(ENG,LE,14.4)
C
C               points of width per point of size
                WPP=E/14.4
C
C               points of size needed for width to be 361 points
                PTS=361.0/WPP
C
C               insert x-coordinate and English in the output line
                CALL STRBNK(PUT,125)
                IF(PTS.GE.14.4) THEN
C                  the text would fit even if bigger than \large
                   W=IFIX(0.5+E)
                   X=(361-W)/2
                   CALL INTINS(PUTE,125,X,6,0, PUT,L)
                   CALL STRINS(PUT,125,ENG,LE,38,0, PUT,L)
                   GO TO 10
                ENDIF
                IF(PTS.GE.12.0) THEN
C                  text won't fit at \large but will at \normalsize
                   W=IFIX(0.5+(12.0/14.4)*E)
                   X=(361-W)/2
                   CALL INTINS(PUTN,125,X,6,0, PUT,L)
                   CALL STRINS(PUT,125,ENG,LE,43,0, PUT,L)
                   GO TO 10
                ENDIF
C               assume the text will fit at \footnotesize
                W=IFIX(0.5+(10.0/14.4)*E)
                X=(361-W)/2
                CALL INTINS(PUTF,125,X,6,0, PUT,L)
                CALL STRINS(PUT,125,ENG,LE,45,0, PUT,L)
C
C               insert the y-coordinate in the output line
   10           CALL INTINS(PUT,125,Y,18,0, PUT,L)
C
C               remove extra blanks and write the output line
                CALL SPACEL(PUT(38),125-38+1)
                L=LENGTH(PUT,125)
                CALL SHIFTL(PUT(L-1),125-(L-1)+1)
                CALL SHIFTL(PUT(18),125-18+1)
                CALL SHIFTL(PUT(6),125-6+1)
                L=LENGTH(PUT,125)
                WRITE(6,903) (PUT(K),K=1,L)
    9      CONTINUE
    4 CONTINUE
C
C     end the back page
      WRITE(6,904)
      RETURN
      END
C
      SUBROUTINE TITLE
C     This routine writes a title page if there is one.
C
C     variable  meaning
C     --------  -------
C     ARG       the command-line parameter as one long word
C     ARGC      ARG as separate characters
C     EXIT      system routine stops program with a return code
C     GETARG    system routine gets one command-line argument
C     IARGC     system function for number of command-line arguments
C     K         index on the characters of ARGC
C     L         length of the title
C     LENGTH    function returns index of last nonblank in a string
C     NPAR      number of command-line parameters found
C
      CHARACTER*24 ARG
      CHARACTER*1 ARGC(24)
      EQUIVALENCE(ARG,ARGC)
C
C ------------------------------------------------------------------
C
C     look for a parameter
      NPAR=IARGC()
      IF(NPAR.EQ.0) RETURN
      IF(NPAR.GT.1) THEN
         WRITE(0,991)
  991    FORMAT('usage: cat flash.in | flashcards "a title"')
         CALL EXIT(3)
      ENDIF
C
C     get the parameter
      CALL GETARG(1,ARG)
      L=LENGTH(ARG,24)
      IF(L.EQ.0) RETURN
C
C     write LaTeX commands to set a title page
      WRITE(6,901) (ARGC(K),K=1,L),'}}'
  901 FORMAT('\vspace*{\fill}'/
     ;       '\centerline{\Large\bf ',25A1)
      WRITE(6,902)
  902 FORMAT('\vspace{\fill}'/
     ;       '\vspace{\fill}'/
     ;       '\newpage')
      RETURN
      END
