C
C     This routine reads from standard in transliterated Hebrew
C     written left-to-right and writes on standard out the
C     corresponding transliterated Hebrew written right-to-left
C     and justified in \hebline{}s.
C
C     unit  assignment
C     ----  ----------
C       0   errors
C       1   ${HOME}/Utility/hebrew.hsh momentarily via HTPWID
C       5   input LaTeX set left-to-right
C       6   output LaTeX set right-to-left in \hebline{}s
C
C     variable  meaning
C     --------  -------
C     ARG       command-line parameter string
C     ERRCH     ERROR as separate characters
C     ERRHB2    error conditions found by HB2HSH
C     ERRHTP    error conditions found by HTPWID
C     ERROR     an error condition
C     FLAG      first 5 characters of an input line, maybe "% LTR"
C     GETARG    system routine gets one command-line argument
C     HEBLNE    \hebline{
C     HEBPNK    \hebpnk{
C     HEBSPC    \hebrew{space}
C     HTPWID    routine separates subtokens in a Hebrew text string
C     I         index on the substrings in a token
C     IARGC     system function for number of command-line arguments
C     IN        input text line
C     INLTR     T => in left-to-right mode
C     JSTART    index in OUT where a token begins
C     K         index on the characters of a string
C     KS        index on the characters of HEBPNK
C     L         length of an error message
C     LENGTH    function returns index of last nonblank in a string
C     LIN       length of an input line
C     LNUM      input file line number just read
C     LS        length of a subtoken
C     LTK       length of a token
C     NEW       T => this is a new output line
C     NST       number of subtokens found in the current token
C     OUT       output text line
C     PNK       a punctuation character
C     PT        point size that will be used for printing
C     RC        return codes; 0 => all went well
C     SHFTKN    routine shifts string left by given number of tokens
C     SPACEL    routine removes leading and extra blanks
C     SUBS      subtokens found in the current token
C     TOKEN     token composed of Hebrew, <transliteration>, punct
C     W         printed width of current output line
C     WPG       width of printed page in points
C     WRTLIN    routine writes OUT and resets it
C     WSP       width of inter-word space at 17 points; see hebwid.f
C     WTK       printed width of the current token
C
C     prepare to turn left-to-right processing on and off
      LOGICAL*4 INLTR/.FALSE./
      CHARACTER*1 IN(1024)
      CHARACTER*5 FLAG
      EQUIVALENCE(IN,FLAG)
C
C     templates
      CHARACTER*1 HEBLNE(9)/'\','h','e','b','l','i','n','e','{'/
      CHARACTER*1 HEBPNK(8)/'\','h','e','b','p','n','k','{'/
      CHARACTER*1 HEBSPC(14)/'\','h','e','b','r','e','w','{','s',
     ;                       'p','a','c','e','}'/
C
C     variables related to width calculations
      CHARACTER*5 ARG
      REAL*4 PT,W,WTK,WSP,WPG/469.0/
C
C     error conditions reported through HTPWID
      CHARACTER*47 ERRHTP(7)/
     ;  'too many subtokens in TOKEN                    ',
     ;  'no closing > on a transliteration              ',
     ;  'transliteration not found in the dictionary    ',
     ;  'no closing } on a \vowel{letter} string        ',
     ;  'HB2HSH returned a hash of more than 1 character',
     ;  'STRWID did not find a character in its tables  ',
     ;  'HEBWID did not find a hash code in its table   '/
      CHARACTER*47 ERRHB2(11)/
     ;  'HSTR has zero length, or MXHSTR=0              ',
     ;  'first character of HSTR is not backslash       ',
     ;  'premature end of HSTR                          ',
     ;  'more than 10 characters in a vowel/trope name  ',
     ;  'vowel/trope name not recognized                ',
     ;  'more than MXHASH vowels/tropes in HSTR         ',
     ;  'more than 10 characters in a letter name       ',
     ;  'letter name not recognized                     ',
     ;  'bad coordinate in a trope specification        ',
     ;  'more than MXHASH letters/{x}{y} pairs in HSTR  ',
     ;  'char after a letter not backslash or space     '/
      CHARACTER*47 ERROR
      CHARACTER*1 ERRCH(47)
      EQUIVALENCE(ERROR,ERRCH)
C
C     other variables
      CHARACTER*1 OUT(1024),TOKEN(1024),SUBS(20,10),PNK
      INTEGER*4 RC
      LOGICAL*4 NEW/.TRUE./
C
C ------------------------------------------------------------------
C
C     get the point size that will be used for printing
      IF(IARGC().EQ.0) GO TO 1
      CALL GETARG(1,ARG)
      READ(ARG,*,ERR=1) PT
      WSP=(555./50.)*(PT/17.0)
      LNUM=0
      GO TO 2
    1 WRITE(0,901)
  901 FORMAT('usage: hebjst pt'/
     ;       '       where pt is REAL*4 point size')
      STOP
C
C     get the next input line
    2 READ(5,902,END=3) IN
  902 FORMAT(1024A1)
      LNUM=LNUM+1
      CALL SPACEL(IN,1024)
      LIN=LENGTH(IN,1024)
C
C     interpret blank lines
      IF(LIN.EQ.0) THEN
C        copy it if outside of L-to-R
         IF(.NOT.INLTR) THEN
            WRITE(6,903)
  903       FORMAT(' ')
            GO TO 2
         ENDIF
C
C        inside L-to-R it signals a paragraph break
         CALL WRTLIN(OUT, JSTART,W,NEW)
         WRITE(6,904)
  904    FORMAT(' '/'\vspace{1\hem}')
         GO TO 2
      ENDIF
C
C     toggle left-to-right processing on and off
      IF(FLAG.EQ.'% LTR') THEN
         IF(INLTR) THEN
C           terminate left-to-right processing
            CALL WRTLIN(OUT, JSTART,W,NEW)
            INLTR=.FALSE.
         ELSE
C           begin left-to-right processing
            INLTR=.TRUE.
            DO 4 K=1,1024
                 OUT(K)=' '
                 IF(K.LE.9) OUT(K)=HEBLNE(K)
                 IF(K.EQ.1024) OUT(K)='}'
    4       CONTINUE
            JSTART=1024-1
            NEW=.TRUE.
         ENDIF
C        skip the flag line itself
         GO TO 2
      ENDIF
C
C     just copy lines that are not to be reversed
      IF(.NOT.INLTR) THEN
         WRITE(6,902) (IN(K),K=1,LIN)
         GO TO 2
      ENDIF
C
C     in L-to-R; extract the input token that comes next
   12 LTK=0
      DO 5 K=1,LIN
           IF(IN(K).EQ.' ') GO TO 6
           LTK=LTK+1
           TOKEN(K)=IN(K)
    5 CONTINUE
C
C     find its width and subtokens
    6 CALL HTPWID(TOKEN,LTK,PT, SUBS,NST,WTK,RC)
      IF(RC.NE.0) THEN
         WRITE(0,905) LNUM,RC,(TOKEN(K),K=1,LTK)
  905    FORMAT('on input line ',I3,' HTPWID returned RC=',I3,
     ;          ' for token ',1024A1)
         IF(RC.LT.0) THEN
C           report the problem detected by HTPWID
            ERROR=ERRHTP(-RC)
         ELSE
C           report the problem detected by HB2HSH
            ERROR=ERRHB2(RC)
         ENDIF
         L=LENGTH(ERRCH,47)
         WRITE(0,906) (ERRCH(K),K=1,L)
  906    FORMAT(47A1)
         STOP
      ENDIF
C
C     will the current output line still fit if we add this token?
      IF(W+WSP+WTK .GT. WPG) THEN
C        no; write the current output line and start a new one
         CALL WRTLIN(OUT, JSTART,W,NEW)
      ENDIF
C
C     space from the previous one any token that is not the first
      IF(.NOT.NEW) THEN
         JSTART=JSTART-14
         DO 7 K=1,14
              OUT(JSTART+K)=HEBSPC(K)
    7    CONTINUE
         W=WSP+W
      ENDIF
C
C     assemble the subtokens in reverse order, R-to-L in output
      DO 8 I=1,NST
           LS=LENGTH(SUBS(1,I),20)
           IF(SUBS(1,I).EQ.'<' .OR. SUBS(1,I).EQ.'\') THEN
C             the subtoken is a transliteration or command
              JSTART=JSTART-LS
              DO 9 K=1,LS
                    OUT(JSTART+K)=SUBS(K,I)
    9         CONTINUE
           ELSE
C             punctuation; \hebpnk its characters one at a time
              DO 10 K=1,LS
                    PNK=SUBS(K,I)
                    JSTART=JSTART-10
                    DO 11 KS=1,8
                          OUT(JSTART+KS)=HEBPNK(KS)
   11               CONTINUE
                    OUT(JSTART+9)=PNK
                    OUT(JSTART+10)='}'
   10         CONTINUE
           ENDIF
    8 CONTINUE
      NEW=.FALSE.
      W=WTK+W
C
C     shift this token out of the input line and go for the next
      CALL SHFTKN(IN,LIN,1,RC)
      LIN=LENGTH(IN,LIN)
      IF(LIN.EQ.0) THEN
C       there are no more tokens; go for the next input line
        GO TO 2
      ELSE
C       there are more tokens in this input line; go for the next
        GO TO 12
      ENDIF
C
C     there are no more input lines
    3 STOP
      END
C
      SUBROUTINE WRTLIN(OUT,JSTART,W,NEW)
C     This routine writes the current output line.
C
C     variable  meaning
C     --------  -------
C     HEBLNE    \hebline{
C     JSTART    index in OUT where a token begins
C     K         index on the characters of OUT
C     LENGTH    function returns index of last nonblank in a string
C     LOUT      length of OUT
C     NEW       T => this is a new output line
C     OUT       output line written and reset
C     SHIFTL    routine removes leading blanks from a string
C     W         printed width of current output line
C
C     formal parameters
      CHARACTER*1 OUT(1024)
      REAL*4 W
      LOGICAL*4 NEW
C
C     template
      CHARACTER*1 HEBLNE(9)/'\','h','e','b','l','i','n','e','{'/
C
C ------------------------------------------------------------------
C
C     finish the current output line and write it
      CALL SHIFTL(OUT(10),1024-9)
      LOUT=LENGTH(OUT,1024)
      WRITE(6,901) (OUT(K),K=1,LOUT)
  901 FORMAT(1024A1)
C
C     start the next output \hebline{}
      DO 1 K=1,1024
           OUT(K)=' '
           IF(K.LE.9) OUT(K)=HEBLNE(K)
           IF(K.EQ.1024) OUT(K)='}'
    1 CONTINUE
      JSTART=1024-1
      W=0.0
      NEW=.TRUE.
      RETURN
      END
