C
C     This program reads input text containing embedded
C     <transliterations> and outputs a .heb file containing a table
C     of the <transliterations> and their English translations.  It
C     can be used to construct a vocabulary list for the input
C     document, to include in another .heb document.
C
C     unit  use
C     ----  --------------------------------------------------------
C       0   standard error: error messages
C       1   ${HOME}/Utility/hebrew.hsh
C       5   standard in: input with embedded <transliterations>
C       6   standard out: output table of unique words
C
C     variable  meaning
C     --------  -------
C     AMP       the string ' & '
C     BLANKS    a string of 18 blanks
C     ENG       translations of the listed words
C     ENGC      ENG columns as individual characters
C     ENOP      translations with (parens) and punctuation removed
C     GETARG    system routine gets one command-line argument
C     HOME      function prepends home directory to a pathname
C     I         index on transliterations
C     IARGC     system function for number of command-line arguments
C     IN        T => in a <transliteration>
C     I4        integer for sorting
C     J         second index on transliterations
C     K         index on the characters of a string
C     KL        index in LINE of leftmost char in a transliteration
C     KR        index in LINE of rightmost char in a transliteration
C     KT        index on the characters of a transliteration
C     L         index in LINE of its last nonblank character
C     LE        length of an English translation
C     LENGTH    function returns index of last nonblank in a string
C     LEXHEB    routine for lexicographical sort of Hebrew hashcodes
C     LINE      a text line read from the input .heb document
C     LS        letter hash codes of transliterations found
C     LT        length of a transliteration
C     MXCH      maximum number of Hebrew characters in a word
C     MXTL      maximum number of transliterations allowed
C     ND        number of duplicate transliterations found
C     NPAR      number of run-line parameters found
C     NT        number of transliterations found
C     ORDER     order in which to print the transliterations
C     ORIG      original order of the transliterations
C     SLASH2    the string '\\'
C     SLASHZ    the hexits for SLASH2
C     SORTI4    routine sorts an INTEGER*4 vector in ascending order
C     T18       an 18-character string for sorting
C     T80       an 80-character string for sorting
C     TL        a transliteration
C     TLS       table of transliterations in input text
C     TLSC      TLS as separate characters
C     UPSTR     routine translates a string to upper case
C     VS        vowel hash codes of transliterations found
C     WAY       how to order the transliterations for output
C     WTOABY    routine removes tokens not to alphabetize by
C
C     prepare to read and parse lines from the input .heb file
      CHARACTER*48 HOME
      CHARACTER*1 LINE(512)
      LOGICAL*4 IN
C
C     prepare to sort the transliterations found there
      PARAMETER(MXTL=256)
      CHARACTER*18 TLS(MXTL),T18,BLANKS/'                  '/
      CHARACTER*1 TLSC(18,MXTL)/4608*' '/
      EQUIVALENCE(TLSC,TLS)
C
C     prepare to store hashcodes and English of the transliterations
      CHARACTER*18 TL
      PARAMETER(MXCH=12)
      INTEGER*1 LS(MXCH,MXTL),VS(MXCH,MXTL)
      CHARACTER*80 ENG(MXTL),ENOP(MXTL)
      CHARACTER*1 ENGC(80,MXTL)
      EQUIVALENCE(ENGC,ENG)
C
C     prepare to output the table
      CHARACTER*2 WAY
      INTEGER*4 ORIG(MXTL)/256*0/,ORDER(MXTL)/256*0/
      CHARACTER*80 T80
      CHARACTER*1 AMP(3)/' ','&',' '/,SLASH2(2)
      INTEGER*1 SLASHZ(2)/Z'5C',Z'5C'/
      EQUIVALENCE(SLASHZ,SLASH2)
C
C ------------------------------------------------------------------
C
C     find out how to sort the transliterations
      NPAR=IARGC()
      IF(NPAR.GT.1) THEN
         WRITE(0,901)
  901    FORMAT('usage: heblist [HE | EH] < fyle.heb')
         STOP
      ENDIF
      IF(NPAR.EQ.0) THEN
         WAY='NO'
      ELSE
         CALL GETARG(1,WAY)
         CALL UPSTR(WAY,2)
         IF(WAY.NE.'HE' .AND. WAY.NE.'EH') THEN
            WRITE(0,901)
            STOP
         ENDIF
      ENDIF
C
C     parse the input text and list all transliterations
      NT=0
      IN=.FALSE.
    2 READ(5,902,END=1) LINE
  902 FORMAT(512A1)
      L=LENGTH(LINE,512)
      IF(L.LT.3) GO TO 2
      IF(LINE(1).EQ.'%') GO TO 2
C
C     extract the transliterations reading right to left
      DO 3 K=L,1,-1
C          find the right end
           IF(LINE(K).EQ.'>') THEN
              IF(IN) THEN
                 WRITE(0,903)
  903            FORMAT('> encountered while in transliteration')
                 STOP
              ELSE
                 IN=.TRUE.
                 KR=K-1
              ENDIF
              GO TO 3
           ENDIF
C          find the left end
           IF(LINE(K).EQ.'<') THEN
              IF(IN) THEN
                 IN=.FALSE.
                 KL=K+1
C                copy transliteration into table, left to right
                 NT=NT+1
                 DO 4 KT=1,KR-KL+1
                       TLSC(KT,NT)=LINE(K+KT)
    4            CONTINUE
              ELSE
                 WRITE(0,904)
  904            FORMAT('< encountered outside transliteration')
                 STOP
              ENDIF
              GO TO 3
           ENDIF
    3 CONTINUE
      GO TO 2
C
C     remember the original order of the transliterations
    1 DO 5 I=1,NT
           ORIG(I)=I
    5 CONTINUE
C
C     make duplicate transliterations blank, and count them
      ND=0
      DO 6 I=1,NT-1
           IF(TLS(I).NE.BLANKS) THEN
              DO 7 J=I+1,NT
                   IF(TLS(J).EQ.TLS(I)) THEN
                      TLS(J)=BLANKS
                      ND=ND+1
                   ENDIF
    7         CONTINUE
           ENDIF
    6 CONTINUE
C
      WRITE(0,905) NT,NT-ND
  905 FORMAT('found',I4,' transliterations '
     ;       'of which',I4,' are unique')
C
C     alphabetize transliterations for 1-pass file search
      DO 8 I=1,NT-1
      DO 8 J=I+1,NT
           IF(TLS(J).LT.TLS(I)) THEN
              T18=TLS(I)
              TLS(I)=TLS(J)
              TLS(J)=T18
              I4=ORIG(I)
              ORIG(I)=ORIG(J)
              ORIG(J)=I4
           ENDIF
    8 CONTINUE
C
C     copy up to discard any blanks from the top of the list
      IF(ND.GT.0) THEN
         DO 9 I=ND+1,NT
              TLS(I-ND)=TLS(I)
              TLS(I)=BLANKS
              ORIG(I-ND)=ORIG(I)
              ORIG(I)=0
    9    CONTINUE
         NT=NT-ND
      ENDIF
C
C     read dictionary extracting data for listed transliterations
      OPEN(UNIT=1,FILE=HOME('/Utility/hebrew.hsh',19))
      I=1
   11 READ(1,906,END=10) TL,(VS(K,I),LS(K,I),K=1,MXCH),
     ;                      (ENGC(K,I),K=1,80)
  906 FORMAT(A18,4X,12(2Z2),1X,80A1)
      IF(TL.EQ.TLS(I)) I=I+1
      IF(I.LE.NT) GO TO 11
C
C     are the transliterations to be alphabetized?
      IF(WAY.EQ.'NO') THEN
C        no; put them back in their original order
         CALL SORTI4(NT,ORIG,ORDER)
         GO TO 12
      ENDIF
C
C     yes; re-sort to table Hebrew->English or English->Hebrew
      DO 13 I=1,NT
            ORDER(I)=I
   13 CONTINUE
      IF(WAY.EQ.'HE') THEN
C        sort in increasing numerical order of the hashcodes
         CALL LEXHEB(MXCH,NT,LS,VS, ORDER)
      ELSE
C        figure out what words to alphabetize by
         DO 14 I=1,NT
               ENOP(I)=ENG(I)
               CALL WTOABY(ENOP(I))
   14    CONTINUE
C
C        sort into alphabetical order of the English
         DO 15 I=1,NT-1
         DO 15 J=I+1,NT
               IF(ENOP(J).LT.ENOP(I)) THEN
                  T80=ENOP(I)
                  ENOP(I)=ENOP(J)
                  ENOP(J)=T80
                  I4=ORDER(I)
                  ORDER(I)=ORDER(J)
                  ORDER(J)=I4
               ENDIF
   15    CONTINUE
      ENDIF
C
C     start the LaTeX table of transliterations
   12 WRITE(6,907)
  907 FORMAT('\begin{figure}[H]'/
     ;       '\centering'/
     ;       '\begin{tabular}{rl}')
C
C     write the contents of the table
      DO 16 I=1,NT
            J=ORDER(I)
            LT=LENGTH(TLSC(1,J),18)
            LE=LENGTH(ENGC(1,J),80)
            IF(WAY.EQ.'HE' .OR. WAY.EQ.'NO') THEN
C              output Hebrew vs English
               WRITE(6,908) '<',(TLSC(KT,J),KT=1,LT),'>',
     ;                      AMP,(ENGC(K,J),K=1,LE),SLASH2
  908          FORMAT(256A1)
            ELSE
C              way is EH; output English vs Hebrew
               WRITE(6,908) (ENGC(K,J),K=1,LE),AMP,
     ;                     '<',(TLSC(KT,J),KT=1,LT),'>',SLASH2
            ENDIF
   16 CONTINUE
C
C     finish the table
      WRITE(6,909)
  909 FORMAT('\end{tabular}'/
     ;       '\end{figure}')
      STOP
C
C     report bad input
   10 WRITE(0,910) TLS(I)
  910 FORMAT('transliteration ',A18,' not found in dictionary')
      STOP
C
C     report too many transliterations
      WRITE(0,911) MXTL
  911 FORMAT('more than ',I2,
     ;       ' transliterations found; increase MXTL')
      STOP
      END
C
      SUBROUTINE WTOABY(ENOP)
C     This routine removes from ENOP tokens not to alphabetize by.
C
C     variable  meaning
C     --------  -------
C     BAD       first words that should not be be alphebetized
C     ENOP      the English string
C     FINDST    routine finds starting location of a substring
C     I         index on bad first words
C     K         index on the characters of ENOP
C     KSTART    index in ENOP where a bad word begins
C     LB        lengths of bad words
C     LE        length of ENOP
C     LENGTH    function returns index of last nonblank in a string
C     NB        number of bad first words
C     SPACEL    routine removes leading and extra blanks
C     UNPUNC    routine removes noise
C
C     formal parameter
      CHARACTER*1 ENOP(80)
C
C     prepare to remove bad extra words
      PARAMETER(NB=13)
      INTEGER*4 LB(NB)/1,3,2,3,2,2,4,2,3,3,4,3,5/
      CHARACTER*20 BAD(NB)/'I                   ',
     ;                     'you                 ',
     ;                     'he                  ',
     ;                     'she                 ',
     ;                     'it                  ',
     ;                     'we                  ',
     ;                     'they                ',
     ;                     'my                  ',
     ;                     'his                 ',
     ;                     'her                 ',
     ;                     'your                ',
     ;                     'our                 ',
     ;                     'their               '/
C
C ------------------------------------------------------------------
C
C     remove punctuation and parenthesized substrings and respace
      CALL UNPUNC(ENOP,80)
C
C     if there is only one token nothing more is to be done
      LE=LENGTH(ENOP,80)
      DO 1 K=1,LE
           IF(ENOP(K).EQ.' ') GO TO 2
    1 CONTINUE
      RETURN
C
C     there is more than one token; remove any bad extra words
    2 DO 3 I=1,NB
           CALL FINDST(ENOP,LE,BAD(I),LB(I), KSTART)
           IF(KSTART.LE.0) GO TO 3
           DO 4 K=KSTART,KSTART+LB(I)
                ENOP(K)=' '
    4      CONTINUE
    3 CONTINUE
      CALL SPACEL(ENOP,LE)
      LE=LENGTH(ENOP,80)
C
      RETURN
      END
