C
C     This program sorts the words in a .dat base dictionary on
C     standard-in into alphabetical order of their transliterations
C     or numerical order of their hashcodes, producing a sorted
C     hashed dictionary on standard-out.
C
C     unit  use
C     ----  -----------
C       0   error messages
C       5   .dat base dictionary
C       6   sorted hashed dictionary
C
C     variable  meaning
C     --------  -------
C     ARG       a command-line parameter
C     ENGS      the English translations of the words
C     EXIT      system routine stops program with a return code
C     FMT       (A18,1X,I2,1X,nn(2I2.2),1X,80A1)
C     GETARG    Unix routine gets a command-line argument
C     HBLIST    routine hashes Hebrew and packs transliteration
C     I         index on words on the dictionary
C     IARGC     Unix routine returns number of command-line args
C     II        index in unsorted file of a record in sorted file
C     K         index on the characters of a string
C     KERR      index in LINE if an error occurs
C     L         number of characters in an English translation
C     LENGTH    function returns index of last nonblank in a string
C     LEXHEB    routine sorts by hashcodes
C     LINE      a dictionary line
C     LS        letter hash codes
C     MXCH      maximum number of Hebrew characters in a word
C     MXWD      maximum number of words in the dictionary
C     NARG      number of command-line arguments found
C     NCH       numbers of characters in the words
C     NWD       number of words found in the dictionary
C     ORDER     indices in unsorted file of sorted records
C     RC        return codes; 0 => all went well
C     SORT      T => sort by transliteration, H => sort by hashcodes
C     TRANS     transliterations
C     TRNSRT    routine sorts by transliteration
C     VS        vowel hash codes
C
C     set maximum sizes
      PARAMETER(MXCH=12,MXWD=6000)
C
C     prepare to sort by either transliteration or hashcodes
      CHARACTER*1 SORT,ARG
      INTEGER*4 ORDER(MXWD)
C
C     prepare to read input lines
      CHARACTER*1 LINE(256)
C
C     prepare to store hashcode lengths and hashcodes
      INTEGER*1 NCH(MXWD),LS(MXCH,MXWD),VS(MXCH,MXWD)
      INTEGER*4 RC
C
C     prepare to store transliterations and translations
      CHARACTER*18 TRANS(MXWD)
      CHARACTER*1 ENGS(80,MXWD)
C
C     prepare to write output lines
      CHARACTER*1 FMT(32)/'(','A','1','8',',','1','X',',','I','2',
     ;                    ',','1','X',',','n','n','(','2','Z','2',
     ;                    '.','2',')',',','1','X',',','8','0','A',
     ;                    '1',')'/
C
C     prepare to report errors
      CHARACTER*59 ERRORS(16)/
     ;'the Hebrew string has zero length                          ',
     ;'the first character is not a backslash                     ',
     ;'string exhausted parsing vowel name                        ',
     ;'vowel name too long                                        ',
     ;'vowel name not found                                       ',
     ;'vowel hash length limit exceeded                           ',
     ;'letter hash length limit exceeded coding a blank           ',
     ;'string exhausted parsing letter name                       ',
     ;'letter name too long                                       ',
     ;'letter name not found                                      ',
     ;'character following a letter name is not backslash or space',
     ;'input line contains only Hebrew text                       ',
     ;'English translation is too long                            ',
     ;'English translation is empty                               ',
     ;'transliteration is too long                                ',
     ;'transliteration is empty                                   '/
C
C ------------------------------------------------------------------
C
C     complete the output format
      CALL BTD(MXCH,FMT(15),2,RC)
C
C     find out whether to sort by transliteration or hashcodes
      SORT='T'
      NARG=IARGC()
      IF(NARG.EQ.1) THEN
         CALL GETARG(1,ARG)
         IF(ARG.EQ.'H' .OR. ARG.EQ.'h') SORT='H'
      ENDIF
C
C     construct the hash and transliteration tables
      NWD=0
    2 READ(5,901,END=1) LINE
  901 FORMAT(256A1)
        NWD=NWD+1
        IF(NWD.GT.MXWD) THEN
           WRITE(0,902) MXWD
  902      FORMAT('found more than ',I4,' input records')
           NWD=MXWD
           GO TO 1
        ENDIF
C
C       process the line
        CALL HBLIST(LINE,MXCH,NCH(NWD),
     ;              LS(1,NWD),VS(1,NWD),ENGS(1,NWD),TRANS(NWD),
     ;              RC,KERR)
        IF(RC.NE.0) THEN
           WRITE(0,904) NWD,KERR,ERRORS(RC)
  904      FORMAT('line ',I4,' column ',I4,':',1X,A59)
           WRITE(0,901) LINE
           WRITE(0,901) (' ',K=1,KERR-1),'|'
           CALL EXIT(0)
        ENDIF
C
C       get the next line
      GO TO 2
C
C     to begin with the lines are in file order
    1 DO 3 I=1,MXWD
           ORDER(I)=I
    3 CONTINUE
C
C     sort the words
      IF(SORT.EQ.'T') THEN
C        into increasing alphabetical order of transliteration
         CALL TRNSRT(NWD, TRANS, ORDER)
C
C        write the sorted transliterations and their hash codes
         DO 4 I=1,NWD
              II=ORDER(I)
              L=LENGTH(ENGS(1,II),80)
              WRITE(6,FMT) TRANS(I),NCH(II),
     ;                     (VS(K,II),LS(K,II),K=1,MXCH),
     ;                     (ENGS(K,II),K=1,L)
    4    CONTINUE
      ELSE
C        into increasing numerical order of hashcodes
         CALL LEXHEB(MXCH,NWD, LS,VS, ORDER)
C
C        write the sorted transliterations and their hash codes
         DO 5 I=1,NWD
              II=ORDER(I)
              L=LENGTH(ENGS(1,II),80)
              WRITE(6,FMT) TRANS(II),NCH(II),
     ;                     (VS(K,II),LS(K,II),K=1,MXCH),
     ;                     (ENGS(K,II),K=1,L)
    5    CONTINUE
      ENDIF
      CALL EXIT(0)
      END
C
      SUBROUTINE HBLIST(LINE,MXCH, NCH,CHASH,VHASH,ENG,TRAN,RC,KERR)
C     This routine hashes the Hebrew word in LINE and packs
C     its transliteration.
C
C     return code RC  meaning
C     --------------  -------
C     1...11          LINE has a syntax error in column KERR
C     12              LINE contains only Hebrew text
C     13              English translation is too long
C     14              English translation is empty
C     15              transliteration is too long
C     16              transliteration empty
C
C     variable  meaning
C     --------  -------
C     CHASH     numbers of the letters in this line
C     ENG       English translation of the word
C     HB2HSH    routine hashes a string of Hebrew text
C     K         index on the characters of LINE
C     KE        index on the characters of ENG
C     KERR      index in LINE where if error is found
C     KSTART    index in LINE where transliteration begins
C     KT        index on the characters of TRAN
C     L         length of LINE
C     LENGTH    function returns index of last nonblank in a string
C     LINE      input string read from millon.dat
C     MXCH      maximum number of characters allowed in a word
C     MXHSTR    length of the Hebrew string in LINE
C     NCH       number of characters in the word
C     RC        return code; see above
C     STRBNK    routine blanks out a character string
C     TRAN      transliteration returned
C     VHASH     numbers of the vowels in this line
C
C     formal parameters
      CHARACTER*1 LINE(256),ENG(80),TRAN(18)
      INTEGER*1 NCH,CHASH(MXCH),VHASH(MXCH)
      INTEGER*4 RC
C
C -----------------------------------------------------------------
C
C     find the end of the Hebrew part of the line
      L=LENGTH(LINE,256)
      DO 1 K=1,L
           IF(LINE(K).EQ.' ') THEN
              MXHSTR=K-1
              GO TO 2
           ENDIF
    1 CONTINUE
      RC=12
      RETURN
C
C     hash the Hebrew
    2 CALL HB2HSH(MXHSTR,LINE,MXCH, NCH,CHASH,VHASH,RC,KERR)
      IF(RC.NE.0) RETURN
C
C     copy the English translation
      KE=0
      CALL STRBNK(ENG,80)
      DO 3 K=MXHSTR+2,L
           KSTART=K+1
           IF(LINE(K).EQ.'%') GO TO 4
           KE=KE+1
           IF(KE.GT.80) THEN
              RC=13
              RETURN
           ENDIF
           ENG(KE)=LINE(K)
    3 CONTINUE
    4 IF(KE.EQ.0) THEN
         RC=14
         KERR=MXHSTR+2
         RETURN
      ENDIF
C
C     copy the transliteration, packing out blanks
      KT=0
      CALL STRBNK(TRAN,18)
      DO 5 K=KSTART,L
           IF(LINE(K).EQ.' ') GO TO 5
           KT=KT+1
           IF(KT.GT.18) THEN
              RC=15
              RETURN
           ENDIF
           TRAN(KT)=LINE(K)
    5 CONTINUE
      IF(KT.EQ.0) THEN
         RC=16
         KERR=KSTART
      ENDIF
      RETURN
      END
C
      SUBROUTINE TRNSRT(NWD, TRANS, ORDER)
C     This routine sorts the transliterations in TRANS into
C     alphabetical order.
C
C     variable  meaning
C     --------  -------
C     I         first index on rows of TRANS
C     ITEMP     temporary for order exchanges
C     J         second index on rows of TRANS
C     NWD       number of rows to sort
C     ORDER     starting order of rows in TRANS, then ending
C     TEMP      temporary for word exchanges
C     TRANS     vector of transliterations to sort

C     formal parameters
      CHARACTER*18 TRANS(NWD)
      INTEGER*4 ORDER(NWD)
C
C     local variables
      CHARACTER*18 TEMP
C
C ------------------------------------------------------------------
C
C     bubble sort
      DO 1 I=1,NWD-1
      DO 1 J=I+1,NWD
C          ascending order is correct
           IF(TRANS(I).LT.TRANS(J)) GO TO 1
C
C          out of order; swap them
           TEMP=TRANS(I)
           TRANS(I)=TRANS(J)
           TRANS(J)=TEMP
C
C          rearrange ORDER to show the new ordering
           ITEMP=ORDER(I)
           ORDER(I)=ORDER(J)
           ORDER(J)=ITEMP
    1 CONTINUE
C
C     report duplicates
      DO 2 I=1,NWD-1
           IF(TRANS(I).EQ.TRANS(I+1)) THEN
              WRITE(0,901) TRANS(I),ORDER(I),ORDER(I+1)
  901         FORMAT(A18,' repeats: ',2I5)
           ENDIF
    2 CONTINUE
      RETURN
      END
