C
C     This program reads LaTeX source text from standard in,
C     replaces each <transliteration> by LaTeX commands to typeset
C     the Hebrew word, and writes the resulting LaTeX source text to
C     standard out.
C
C     unit  use
C     ----  ---
C        1  input dictionary ${HOME}/Utility/hebrew.hsh
C        5  input LaTeX source text containing <transliterations>
C        6  output LaTeX source text containing Hebrew commands
C        0  output error messages
C
C       RC  meaning
C       --  -------
C        0  all went well
C        1  one or more transliterations were not in the dictionary
C        2  mismatched <> delimiters
C        3  transliteration too long
C        4  HSH2HB reported an incorrect hash code
C        5  an \hbf{ opened before a previous \hbf{ closed
C        6  an \hbf{ opened but never closed
C        7  a sectioning command was left unclosed
C        8  there was not enough space to insert a needed \protect
C
C     variable  meaning
C     --------  -------
C     EXIT      system routine stops program with a return code
C     EXPAND    routine replaces transliterations by LaTeX
C     HBF       routine replaces \hbf{} constructs by LaTeX commands
C     K         index on the characters of a string
C     LENGTH    function returns index of last nonblank in a string
C     LINE      LaTeX source text line read and written
C     LL        index of last nonblank in LINE
C     LMAX      dimensioned length of LINE
C     LNUM      current line number in LaTeX input file
C     NLBC      net number of {'s remaining in sectioning command
C     NLBF      net number of {'s remaining in \hbf{}
C     PROTCT    routine \protects Hebrew in sectioning commands
C     RC        return code; see table above
C     RCE       return code from EXPAND
C
C     prepare to read, expand, and write LaTeX source lines
      PARAMETER(LMAX=2048)
      CHARACTER*1 LINE(LMAX)
C
C     prepare to report failure
      INTEGER*4 RC,RCE
C
C ------------------------------------------------------------------
C
C     initialize
      LNUM=0
      NLBF=0
      NLBC=0
      RCE=0
C
C     read the next line of LaTeX input text
    2 READ(5,900,END=1) LINE
  900 FORMAT(2048A1)
      LNUM=LNUM+1
      LL=LENGTH(LINE,2048)
C
C     copy blank lines and comment lines unchanged
      IF(LL.EQ.0) THEN
         WRITE(6,900)
         GO TO 2
      ENDIF
      IF(LINE(1).EQ.'%') THEN
         WRITE(6,900) (LINE(K),K=1,LL)
         GO TO 2
      ENDIF
C
C     expand \hbf{} designations into LaTeX
      CALL HBF(LINE,LL,LNUM, NLBF, RC)
      IF(RC.GT.0) GO TO 1
C
C     expand transliterations into LaTeX
      CALL EXPAND(LINE,LL,LNUM, RC)
      RCE=MAX0(RC,RCE)
      IF(RC.GT.1) GO TO 1
C
C     protect each Hebrew letter in command arguments
      CALL PROTCT(LINE,LMAX, NLBC,LL, RC)
      IF(RC.GT.7) GO TO 1
C
C     write the substituted line
      WRITE(6,900) (LINE(K),K=1,LL)
      GO TO 2
C
C     the end of the input LaTeX file has been reached
    1 IF(NLBF.GT.0) THEN
         RC=6
         WRITE(0,901)
  901    FORMAT('An \hbf{} construct was left unclosed.')
      ENDIF
      IF(NLBC.GT.0) THEN
         RC=7
         WRITE(0,902)
  902    FORMAT('A sectioning command was left unclosed.')
      ENDIF
      RC=MAX0(RC,RCE)
      CALL EXIT(RC)
      END
C
      SUBROUTINE HBF(LINE,LL,LNUM, NLBF, RC)
C     This routine expands \hbf{} into \renewcommand commands.
C
C     variable  meaning
C     --------  -------
C     FINDST    routine finds starting location of a substring
C     FINISH    command to stop bold
C     K         index on the characters of LINE
C     KSTART    index in LINE where a string begins, or 0
C     LINE      LaTeX source text line to process
C     LL        length of LINE
C     NLBF      net number of left braces remaining in \hbf{}
C     RC        return code from this routine; see table in main
C     SHIFTR    routine makes string flush right in its length
C     START     command to start bold
C     STRINS    routine inserts one character string into another
C
C     formal parameters
      CHARACTER*1 LINE(2048)
      INTEGER*4 RC
C
C     prepare to insert boldness-changing commands
      CHARACTER*34 START/'\renewcommand{\hbold}[1]{\pmb{#1}}'/
      CHARACTER*28 FINISH/'\renewcommand{\hbold}[1]{#1}'/
C
C ------------------------------------------------------------------
C
C     assume there will not be an error
      RC=0
C
C     search for the start of a bold string
    3 CALL FINDST(LINE,LL,'\hbf{',5, KSTART)
      IF(KSTART.LE.0 .AND. NLBF.EQ.0) THEN
C        no bold string starts on this line, not looking for an end
         RETURN
      ENDIF
      IF(KSTART.LE.0 .AND. NLBF.GT.0) THEN
C        no bold string starts on this line but looking for an end
         GO TO 1
      ENDIF
      IF(KSTART.GT.0 .AND. NLBF.GT.0) THEN
C        looking for an end but found another beginning
         WRITE(0,901) LNUM,KSTART
  901    FORMAT('LaTeX line ',I4,' column ',I4,':',
     ;          ' \hbf{ begins before previous \hbf{ ends')
         RC=5
         RETURN
      ENDIF
      IF(KSTART.GT.0 .AND. NLBF.EQ.0) THEN
C       found the start of a bold string
        NLBF=1
C
C       open a space for the command
        CALL SHIFTR(LINE(KSTART+5),2048-(KSTART+4))
C
C       insert the command to set bold type
        CALL STRINS(LINE,2048,START,34,KSTART,1, LINE,LL)
      ENDIF
C
C     search for the end of a bold string
    1 DO 2 K=1,LL
           IF(LINE(K).EQ.'{') NLBF=NLBF+1
           IF(LINE(K).EQ.'}') NLBF=NLBF-1
           IF(NLBF.EQ.0) THEN
C             open a space for the command to set normal type
              LINE(K)=' '
              CALL SHIFTR(LINE(K),2048-(K+1))
C
C             insert the command to set normal type
              CALL STRINS(LINE,2048,FINISH,28,K,1, LINE,LL)
              GO TO 3
           ENDIF
    2 CONTINUE
C
C     still looking for an end; check for an erroneous other \hbf
      CALL FINDST(LINE,LL,'\hbf{',5, KSTART)
      IF(KSTART.LE.0) RETURN
      WRITE(0,901) LNUM,KSTART
      RC=5
      RETURN
      END
C
      SUBROUTINE EXPAND(LINE,LL,LNUM, RC)
C     This routine expands <transliterations> into LaTeX.
C
C     variable  meaning
C     --------  -------
C     BLANKS    the string '                '
C     FINDST    routine finds starting location of a substring
C     FIRST     T => this is the first invocation of this routine
C     GETHSH    routine reads in the hashed Hebrew dictionary
C     HBOX      T => encapsulate insertion in \hbox{ }
C     HSH2HB    routine expands the hash of a Hebrew string
C     I         index on dictionary entries
C     IONE0     an INTEGER*1 0
C     IONE1     an INTEGER*1 1
C     IONE38    an INTEGER*1 38
C     K         index on the characters of a string
C     KEND      index in LINE of last character in insertion
C     KPLUS     characters between <delimiters> plus 1
C     KSTART    index in LINE of < of a transliteration
C     KSTOP     index in LINE of > of a transliteration
C     LCHG      net change in LL from processing this <>
C     LHB       number of characters in Hebrew insertion
C     LINE      LaTeX source text line read and written
C     LL        length of LINE
C     LNUM      current line number in LaTeX input file
C     LR        unused; index of last nonblank in LINE
C     LS        each column hash codes letters of a dictionary word
C     LTR       a letter
C     MVDRT     characters moved right to open space for insertion
C     MXCH      allowed number of characters in a dictionary word
C     MXHB      space available for Hebrew insertion
C     MXWD      allowed number of dictionary words
C     NS        character counts of Hebrew words
C     NWD       number of dictionary words found
C     RC        return code from this routine; see table in main
C     RCL       return code from HSH2HB; 0 => all went well
C     SHIFTL    routine removes leading blanks from a string
C     SHIFTR    routine makes string flush right in its length
C     STRINS    routine inserts one string into another
C     TRAN      the transliterations of the dictionary words
C     TRN       the transliteration of a Hebrew word to be inserted
C     TRNC      TRN as separate characters
C     TRNIDX    routine returns index in TRAN of TRN
C     VS        each column hash codes vowels in a dictionary word
C     VWL       a vowel
C
C     prepare to store the file line that is being processed
      CHARACTER*1 LINE(*)
C
C     prepare to store the hashed dictionary, each word in a column
      PARAMETER(MXCH=12,MXWD=6000)
      INTEGER*1 NS(MXWD),LS(MXCH,MXWD),VS(MXCH,MXWD)
      CHARACTER*18 TRAN(MXWD)
C
C     prepare to extract a <transliteration> from LaTeX source
      CHARACTER*18 TRN,BLANKS/'                '/
      CHARACTER*1 TRNC(18)
      EQUIVALENCE(TRN,TRNC)
      INTEGER*1 VWL(18),LTR(18),IONE0/0/,IONE1/1/,IONE38/38/
C
C     other variables
      LOGICAL*4 FIRST/.TRUE./
      INTEGER*4 RCL,RC,TRNIDX
      LOGICAL*4 HBOX
C
C ------------------------------------------------------------------
C
C     read in the hashed dictionary on the first call only
      IF(FIRST) THEN
         FIRST=.FALSE.
         CALL GETHSH(MXCH,MXWD, NWD,NS,TRAN,VS,LS)
      ENDIF
C
C     initialize for the processing of this line
      RC=0
      LCHG=0
C
C     find the next <transliteration> in this input line
    7 CALL FINDST(LINE,LL,'<',1, KSTART)
      IF(KSTART.EQ.0) RETURN
      CALL FINDST(LINE(KSTART+1),LL-KSTART,'>',1, KPLUS)
      IF(KPLUS.EQ.0) THEN
         WRITE(0,902) LNUM,KSTART-LCHG,(LINE(K),K=1,LL)
  902    FORMAT('LaTeX line ',I4,':',I4,' mismatched <>'/2048A1)
         RC=2
         RETURN
      ENDIF
      KSTOP=KSTART+KPLUS
C
C     if there is a blank space after <transliteration> preserve it
      IF(LINE(KSTOP+1).EQ.' ') LINE(KSTOP+1)='#'
C
C     if the <transliteration> stands alone it will be encapsulated
      HBOX=.TRUE.
      IF(KSTART.GT.1) THEN
         IF(LINE(KSTART-1).NE.' ') HBOX=.FALSE.
      ENDIF
      IF(LINE(KSTOP+1).NE.'#') HBOX=.FALSE.
C
C     extract the transliteration up to the closing >
      TRN=BLANKS
      IF(KSTOP-KSTART-1.GT.18) THEN
         WRITE(0,905) LNUM
  905    FORMAT('transliteration too long at input line',I4)
         RC=3
         RETURN
      ENDIF
      DO 4 K=KSTART,KSTOP
           IF(K.GT.KSTART .AND. K.LT.KSTOP) TRNC(K-KSTART)=LINE(K)
C          blanking out the characters as you go
           LINE(K)=' '
    4 CONTINUE
C
C     translate the transliteration into byte codes
      IF(TRN.EQ.BLANKS) THEN
C        the transliteration is blanks
         I=0
         DO 9 K=1,MXCH
              IF(K.LT.(KSTOP-KSTART)) THEN
                 VWL(K)=IONE1
                 LTR(K)=IONE38
              ELSE
                 VWL(K)=IONE0
                 LTR(K)=IONE0
              ENDIF
    9    CONTINUE
         GO TO 8
      ELSE
C        bisect to find the transliteration in the dictionary
         I=TRNIDX(TRN,TRAN,NWD)
         IF(I.EQ.0) THEN
            WRITE(0,903) TRN,LNUM,KSTART-LCHG,KSTOP-LCHG
  903       FORMAT(A18,' not found line',I4,' columns ',I3,'-',I3)
            RC=1
            GO TO 7
         ENDIF
         DO 10 K=1,MXCH
               VWL(K)=VS(K,I)
               LTR(K)=LS(K,I)
   10    CONTINUE
      ENDIF
C
C     open a space for the Hebrew
    8 CALL SHIFTR(LINE(KSTOP+1),2048-(KSTOP+1)+1)
      MVDRT=LL-KSTOP
C
C     insert \hbox{ to encapsulate the Hebrew
      IF(HBOX) THEN
         CALL STRINS(LINE,2048,'\hbox{',6,KSTART,0, LINE,LR)
         KSTART=KSTART+6
      ENDIF
C
C     insert the Hebrew
      MXHB=2048-(KSTART-1)-MVDRT
      CALL HSH2HB(MXCH,LTR,VWL,MXHB, LHB,LINE(KSTART),RCL)
      IF(RCL.NE.0) THEN
         WRITE(0,904) TRAN(I),RCL
  904    FORMAT('the hash of ',A18,' is wrong at letter ',I2)
         RC=4
         RETURN
      ENDIF
C
C     insert } to close the encapsulation
      IF(HBOX) THEN
         LINE(KSTART+LHB)='}'
         LHB=LHB+1
      ENDIF
C
C     close the space
      KEND=KSTART+LHB-1
      CALL SHIFTL(LINE(KEND+1),2048-KEND)
      IF(LINE(KEND+1).EQ.'#') LINE(KEND+1)=' '
C
C     adjust the length of the line and continue
      LCHG=LCHG+(KEND+MVDRT-LL)
      LL=KEND+MVDRT
      GO TO 7
      END
C
      SUBROUTINE PROTCT(LINE,LMAX, NLBC,LL, RC)
C     This routine inserts \protect before each Hebrew vowel
C     in the argument of these commands:
C       \chapter{*} \section{*} \subsection{*}
C
C     variable  meaning
C     --------  -------
C     CMD       LaTeX sectioning command names
C     FINDST    routine finds starting location of a substring
C     HEB       Hebrew vowel names
C     I         index on names
C     IC        index of command name found
C     IH        index of Hebrew vowel name found
C     K         index in LINE or LINE(KCS) or LINE(KHS)
C     KC        index in LINE of \ of sectioning command
C     KCE       index in LINE to end search for command name
C     KCS       index in LINE to start a search for command name
C     KH        index in LINE of \ of Hebrew vowel name
C     KHS       index in LINE to start search for Hebrew vowel
C     LC        lengths of command names
C     LH        lengths of Hebrew vowel names
C     LINE      LaTeX source text line being processed
C     LL        index of last nonblank character in LINE
C     LMAX      dimensioned size of LINE
C     NC        number of command names
C     NH        number of Hebrew vowel names
C     NLBC      net number of left braces remaining in command
C     RC        return code; see table in main program
C     SHIFTR    routine makes string flush right in its length
C     STRINS    routine inserts one character string into another
C
C     formal parameters
      CHARACTER*1 LINE(LMAX)
      INTEGER*4 RC
C
C     command names
      PARAMETER(NC=3)
      CHARACTER*10 CMD(NC)/'chapter   ',
     ;                     'section   ',
     ;                     'subsection'/
      INTEGER*4 LC(NC)/7,7,10/
C
C     possible Hebrew vowel names
      PARAMETER(NH=16)
      CHARACTER*10 HEB(NH)/'hebrew    ',
     ;                     'chiriq    ',
     ;                     'tsere     ',
     ;                     'segol     ',
     ;                     'halfsegol ',
     ;                     'patach    ',
     ;                     'halfpatach',
     ;                     'qamats    ',
     ;                     'awe       ',
     ;                     'qubbuts   ',
     ;                     'sheva     ',
     ;                     'lcholam   ',
     ;                     'rcholam   ',
     ;                     'dagesh    ',
     ;                     'meteg     ',
     ;                     'hebspc    '/
      INTEGER*4 LH(NH)/6,6,5,5,9,6,10,6,3,7,5,7,7,6,5,6/
C
C ------------------------------------------------------------------
C
C     assume LINE is long enough for the needed insertions
      RC=0
C
C     are we in a sectioning command fom the previous input line?
      IF(NLBC.GT.0) GO TO 1
C
C     not in a sectioning command; look for one on this line
      KCS=1
      KC=LL
      IC=0
      DO 2 I=1,NC
           CALL FINDST(LINE(KCS),LL-KCS+1,CMD(I),LC(I), K)
           IF(K.LE.1) GO TO 2
           IF(LINE(KCS+K-2).NE.'\') GO TO 2
C
C          found one of the command names
           IF(K.LT.KC) THEN
C             remember where we found the leftmost one
              KC=K
              IC=I
           ENDIF
    2 CONTINUE
      IF(IC.EQ.0) RETURN
C     make KC be an index in LINE instead of LINE(KCS)
      KC=KCS+KC-2
C
C     in a sectioning command; look for the end of its argument
    1 DO 3 K=KC+1+LC(IC),LL
           IF(LINE(K).EQ.'{') NLBC=NLBC+1
           IF(LINE(K).EQ.'}') NLBC=NLBC-1
           IF(NLBC.EQ.0) THEN
              KCE=K
              GO TO 4
           ENDIF
    3 CONTINUE
      KCE=LL
C
C     if some argument is already \protected, leave the line alone
    4 KHS=KC+1+LC(IC)+1
      CALL FINDST(LINE(KHS),KCE-KHS+1,'\protect',8, K)
      IF(K.GT.0) RETURN
C
C     search for the next Hebrew vowel in the command argument
    6 KH=KCE-KHS
      IH=0
      DO 5 I=1,NH
           CALL FINDST(LINE(KHS),KCE-KHS+1,HEB(I),LH(I), K)
           IF(K.LE.0) GO TO 5
           IF(LINE(KHS+K-2).NE.'\') GO TO 5
C
C          found one of the vowel names
           IF(K.LT.KH) THEN
C             remember where we found the leftmost one
              KH=K
              IH=I
           ENDIF
    5 CONTINUE
      IF(IH.EQ.0) RETURN
C     make KH be an index in LINE instead of LINE(KHS)
      KH=KHS+KH-2
C
C     found a Hebrew vowel; open a space for \protect
      IF(LMAX-LL .LT. 8) THEN
         RC=8
         RETURN
      ELSE
         CALL SHIFTR(LINE(KH),LMAX-KH+1)
      ENDIF
C
C     insert the \protect and close the space
      CALL STRINS(LINE,LMAX,'\protect',8,KH,2, LINE,LL)
      KCE=KCE+8
C
C     continue the search for Hebrew vowels
      KHS=KH+8+LH(IH)+1
      GO TO 6
      END
